/* $Header$ */	    
/*
 *
 *                 Copyright 1994-1996 by Borland Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code
 *    remains the property of Borland Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Borland Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Borland Corp.
 *
 *    Owner:
 *        Borland Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

/* include files */
#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN    /* defines the LibMain() */

#define end_of_line(ch)\
(( ((ch) == EOL1_CHAR)||( (ch) == EOL2_CHAR) || ( (ch) == EOF_CHAR ) )\
? (TRUE) : (FALSE))


#define MAX_BUF_LEN    100
static char buf[MAX_BUF_LEN + 1];

static LPVOID _tex_hash_names        = NIL; /* pointer to the hash table containing */
                                            /* all 'TEX' keyword names. */

static LPVOID _tex_hash_names_1      = NIL; /* pointer to optional hash table containing */
static int _tex_hash_size_1          = 47;  /* user defined keyword names and size of
                                            /* table. When created these key
                                            /* are set to Color Alternate5 */           
                        

static LPVOID _tex_hash_names_2      = NIL; /* pointer to optional hash table containing */
static int _tex_hash_size_2          = 47;  /* user defined keyword names and size of
                                            /* table. When created these key
                                            /* are set to Color Alternate6 */              

static LPVOID _tex_hash_templates    = NIL; /* pointer to the hash table containing */
                                            /* an abbreviated name and its template */



/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _tex_hash_templates.
 */
typedef struct _LanguageTemplate {         
    LPSTR abbrev;
    LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR TEX_INDENT_STR = "\377\\\\begin{enumerate}\377\\\\begin{itemize}\377";



/**************************************************************************
 * _tex_templates[] contains the keyword abbreviation and the template for 
 *  the abbreviation.  To add a new name, insert it before the NIL entry or
 *  call _tex_assign_template().
 *
 * The special characters within the template are:
 *
 *    \n - new line
 *     & - cursor position after template insertion
 *     @ - backspace
 *    \c - insert 'c' literally
 **************************************************************************/
LanguageTemplate _tex_templates[] = {
//    Abbreviation                       Template
    { "\\ac|\\acu|\\acut",			"%10acute"},
    { "\\al|\\alp|\\alph" ,         "%10alpha" },
    { "\\ale|\\alep",				"%10aleph"},
	{ "\\am|\\ama|\\amal", 			"%10amalg"},
	{ "\\an|\\ang|\\angl",			"%10angle"},
	{ "\\ap|\\app|\\appr|\\appro",	"%10approx"},
	{ "\\as|\\asy|\\asym",			"%10asymp"},
	{ "\\ba|\\bac|\\back|\\backs",	"%10backslash"},
    { "\\be|\\bet" ,				"%10beta"  },
	{ "\\beg|\\begi|\\begin",		"%10begin{&}\n\n\\\\end{}"},
	{ "\\begin|\\begin{",       	"%10begin{&}\n\n\\\\end{}" },
	{ "\\begin{enumerate",      	"%10begin{enumerate}\n\t\\\\item&\n@ \\\\end{enumerate}" },
	{ "\\begin{itemize",        	"%10begin{itemize}\n\t\\\\item&\n@ \\\\end{enumerate}" },
	{ "\\begin{center",	        	"%10begin{center}\n&\n\\\\end{center}" },
	{ "\\begin{quote",	        	"%10begin{quote}\n&\n\\\\end{quote}" },
	{ "\\begin{figure",	        	"%10begin{figure}[htb]\n&\n\\\\end{figure}" },
	{ "\\begin{table",	        	"%10begin{table}[htb]\n&\n\\\\end{table}" },
	{ "\\begin{array",	        	"%10begin{array}{&}\n\n\\\\end{array}" },
	{ "\\begin{tabular",        	"%10begin{tabular}{&}\n\n\\\\end{tabular}" },
	{ "\\begin{minipage",       	"%10begin{minipage}[t]{&}\n\n\\\\end{minipage}" },
	{ "\\begin{equation",       	"%10begin{equation}\n&\n\\\\end{equation}" },
	{ "\\begin{eqnarray",	        "%10begin{eqnarray}\n&\n\\\\end{eqnarray}" },
	{ "\\begin{eqnarray*",      	"%10begin{eqnarray}\n&\n\\\\end{eqnarray}" },
	{ "\\[",  			         	"@ %10[\n\t&\n@ %10]"},
	{ "\\begin{letter",	        	"%10begin{letter}\n\t{&\n}\n\n@ %10opening{}\n\n%10closing{%10 }\n%10end{letter}"},					
	{ "\\bibi|\\bibit|\\bibite",	"%10bibitem"},
	{ "\\bibliog",					"%10bibliography"},
	{ "\\bibliogs",					"%10bibliographystyle"},
	{ "\\bigc",						"%10bigcirc"},
	{ "\\bigtd",					"%10bigtriangleup"},
	{ "\\bigtu",					"%10bigtriangledown"},
	{ "\\boldm",					"%10boldmath"},
	{ "\\bowt",						"%10bowtie"},
	{ "\\br|\\bre|\\brev",			"%10breve"},
	{ "\\bu|\\bul|\\bull|\\bulle",	"%10bullet"},
	{ "\\cd|\\cdo",					"%10cdot"},
	{ "\\cir", 						"%10circ"},
	{ "\\cit", 						"%10cite"},
	{ "\\club", 					"%10clubsuit"},
	{ "\\co|\\con", 				"%10cong"},
	{ "\\cop|\\copy|\\copyr", 		"%10copyright"},
	{ "\\dag", 						"%10dagger"},
	{ "\\das|\\dash", 				"%10dashv"},
	{ "\\dat", 						"%10date"},
	{ "\\ddag", 					"%10ddagger"},
	{ "\\ddo", 						"%10ddot"},
	{ "\\De|\\Del|\\Delt" ,         "%10Delta" },
    { "\\de|\\del|\\delt" ,         "%10delta" },
    { "\\Di|\\Dia|\\Diam|\\Diamo|\\Diamon" , 	"%10Diamond" },
	{ "\\dia|\\diam|\\diamo|\\diamon", 			"%10diamond"},
	{ "\\diams", 					"%10diamondsuit"},
    { "\\Downa|\\Downar|\\Downarr|\\Downarro",	"%10Downarrow"},
	{ "\\doc", 						"%10documentstyle"},
	{ "\\dot|\\dote",				"%10doteq"},
	{ "\\downa|\\downar|\\downarr|\\downarro",	"%10downarrow"},
	{ "\\emp|\\empt|\\empty|\\emptys", 			"%10emptyset"},
	{ "\\en|\\end", 				"%10end{&}"},
    { "\\ep|\\eps|\\epsi|\\epsil|\\epsilo" ,	"%10epsilon" },
	{ "\\eqn", 						"%10eqnarray"},
	{ "\\equ", 						"%10equiv"},
	{ "\\even", 					"%10evensidemargin"},
	{ "\\ex|\\exi|\\exis|\\exist",	"%10exists"},
	{ "\\fl|\\fla|",				"%10flat"},
	{ "\\fl|\\flo|\\floo", 			"%10lfloor"},
	{ "\\fora|\\foral",				"%10forall"},
	{ "\\fro|\\frow", 				"%10frown"},
    { "\\ga|\\gam|\\gamm" ,         "%10gamma" },
    { "\\Ga|\\Gam|\\Gamm",          "%10Gamma"},
	{ "\\gra|\\grav", 				"%10grave"},
	{ "\\hb|\\hba", 				"%10hbar"},
	{ "\\headi|\\headin|\\heading",		"%10headings"},
	{ "\\headh|\\headhe|\\headhei|\\headheig|\\headheigh", "%10headheight"},
	{ "\\heads|\\headse",			"%10headsep"},
	{ "\\hearts",					"%10heartsuit"},
	{ "\\hookl", 					"%10hookleftarrow"},
	{ "\\hookr", 					"%10hookrightarrow"},
	{ "\\ima|\\imat", 				"%10imath"},
	{ "\\inf|\\inft", 				"%10infty"},
	{ "\\io|\\iot", 				"%10iota"},
	{ "\\it|\\ite", 				"%10item"},
	{ "\\jm", 						"%10jmath"},
    { "\\Jo|\\Joi",                 "%10Join"},
	{ "\\ka|\\kap|\\kapp", 			"%10kappa"},
	{ "\\la|\\lam|\\lamb|\\lambd",	"%10lambda"},
    { "\\La|\\Lam|\\Lamb|\\Lamda",  "%10Lambda"},
	{ "\\lang",						"%10langle"},
	{ "\\lc",						"%10lceil"},
	{ "\\lefta",					"%10leftarrow"},
    { "\\Lefta|\\Leftar|\\Leftarr|\\Leftarro", "%10Leftarrow"},
	{ "\\lefthd",					"%10leftharpoondown"},
	{ "\\lefthu", 					"%10leftharpoonup"},
    { "\\Leftrighta", 				"%10Leftrightarrow" },
	{ "\\longla", 					"%10longleftarrow"},
    { "\\Longlefta", 				"%10Longleftarrow"},
    { "\\Longleftr",				"%10Longleftrightarrow"},
	{ "\\longm", 					"%10longmapsto"},
	{ "\\longra", 					"%10longleftrightarrow"},
	{ "\\longra", 					"%10longrightarrow"},
    { "\\Longrighta", 				"%10Longrightarrow"},
	{ "\\make|\\maket|\\maketi|\\maketit|\\maketitl", "%10maketitle"},
	{ "\\maps|\\mapst", 			"%10mapsto"},
	{ "\\mb|\\mbo|\\mbox", 			"%10mbox{&}"},
	{ "\\mod|\\mode|\\model",		"%10models"},
	{ "\\na|\\nab|\\nabl",			"%10nabla"},
	{ "\\nea",						"%10nearrow"},
	{ "\\new|\\newc", 				"%10newcommand"},
	{ "\\nw|\\nwa|\\nwar|\\nwarr|\\wnarro", "%10nwarrow"},
	{ "\\od",						"%10odot"},
	{ "\\odd",						"%10oddsidemargin"},
    { "\\Om|\\Ome|\\Omeg", 			"%10Omega"},
	{ "\\om|\\ome|\\omeg|\\omeg", 	"%10omega"},
	{ "\\omin|\\ominu",				"%10ominus"},
	{ "\\op|\\opl|\\oplu",			"%10oplus"},
	{ "\\os|\\osl|\\osla|\\oslas", 	"%10oslash"},
	{ "\\ot|\\oti|\\otim|\\otime", 	"%10otimes"},
	{ "\\pageb", 					"%10pagebreak"},
	{ "\\pagen", 					"%10pagenumbering"},
	{ "\\pages", 					"%10pagestyle"},
	{ "\\par|\\para|\\paral|\\parall|\\paralle", 	"%10parallel"},
	{ "\\pe|\\per", 				"%10perp"},
	{ "\\pr|\\pre", 					"%10prec"},
	{ "\\prec|\\prece", 			"%10preceq"},
	{ "\\pri|\\prim", 				"%10prime"},
	{ "\\ra|\\ran|\\rang|\\rangl",	"%10rangle"},
	{ "\\rc|\\rce|\\rcei",			"%10rceil"},
	{ "\\rf|\\rfl|\\rflo|\\rfloo",	"%10rfloor"},
    { "\\Righta", 					"%10Rightarrow"},
	{ "\\righta", 					"%10rightarrow"},
	{ "\\righthd", 					"%10rightharpoondown"},
	{ "\\righthp", 					"%10rightharpoonup"},
	{ "\\sa|\\sar|\\sarr|\\sarro", 	"%10swarrow"},
    { "\\sec|\\sect|\\secti|\\sectio", "%10section{&}" },
    { "\\set|\\setl|\\setle|\\setlen|\\setleng|\\setlengh", "%10setlenght" },
    { "\\Si|\\Sig|\\Sigm", 			"%10Sigma"},
    { "\\sub|\\subs|\\subse|\\subsec", "%10subsection{&}" },
    { "\\subsub", 					"%10subsubsection{&}" },
    { "\\suc|\\succ|\\succe", 		"%10succeq"},
	{ "\\sups|\\supse", 			"%10supset"},
	{ "\\supsq", 					"%10supseteq"},
	{ "\\sur", 						"%10surd"},
    { "\\tab|\\tabl|\\table", 		"%10tableofcontents" },
	{ "\\texth", 					"%10textheight"},
	{ "\\textw", 					"%10textwidth"},
    { "\\Th|\\The|\\Thet", 			"%10Theta"},
	{ "\\th|\\the|\\thet",			"%10theta"},
	{ "\\theb", 					"%10thebibliography"},
	{ "\\thisp", 					"%10thispagestyle"},
	{ "\\til|\\tild", 			    "%10tilde"},
	{ "\\tit|\\titl", 				"%10title"},
	{ "\\topm", 					"%10topmargin"},
	{ "\\tri", 						"%10triangle"},
	{ "\\tril", 					"%10triangleleft"},
	{ "\\trir", 					"%10triangleright"},
	{ "\\unl|\\unlh", 				"%10unlhd"},
	{ "\\unr|\\unrh", 				"%10unrhd"},
	{ "\\up|\\ups|\\upsi|\\upsil|\\upsilo", 	"%10upsilon"},
	{ "\\upa", 						"%10uparrow"},
    { "\\Upa|\\Upar|\\Uparr|\\Uparro", 			"%10Uparrow"},
    { "\\Upd|\\Updo|\\Updow|\\Updown|\\Updowna", 	"%10Updownarrow"},
	{ "\\upda", 					"%10updownarrow"},
    { "\\Ups|\\Upsi|\\Upsil|\\Upsilo", 			"%10Upsilon"},
	{ "\\vare",	 					"%10varepsilon"},
	{ "\\varp", 					"%10varpi"},
	{ "\\varr", 					"%10varrho"},
	{ "\\vars", 					"%10varsigma"},
	{ "\\vart", 					"%10vartheta"},
	{ "\\vd|\\vda", 				"%10vdash"},
	{ "\\we|\\wed|\\wedg", 			"%10wedge"},
	{ "\\ze|\\zet", 				"%10zeta"},
    { NIL,                          NIL },
};

/**************************************************************************
 * _tex_names[] - contains the list of 'TEX' keyword names which appear 
 *  as "keyword color" when language coloring is enabled.
 *  To add a new keyword, insert it before the NIL entry.
 *************************************************************************/
LPSTR _tex_names[] = {
"\\Box", 
"\\Delta", 
"\\Diamond", 
"\\Downarrow", 
"\\Gamma", 
"\\Im",
"\\Join", 
"\\Lambda", 
"\\Leftarrow", 
"\\Leftrightarrow",
"\\Longleftarrow", 
"\\Longleftrightarrow",
"\\Longrightarrow", 
"\\NeedsTeXFormat",
"\\NEWdescription",
"\\NEWfeature",
"\\Omega", 
"\\Phi", 
"\\Pi", 
"\\Psi", 
"\\Re", 
"\\Rightarrow", 
"\\Sigma", 
"\\Theta", 
"\\Uparrow", 
"\\Updownarrow", 
"\\Upsilon",
"\\Xi", 
"\\acute",
"\\aleph",
"\\alpha", 
"\\amalg", 
"\\angle",
"\\approx", 
"\\arg",
"\\arg",
"\\ast", 
"\\asymp",
"\\author", 
"\\backmatter",
"\\backslash",
"\\bar", 
"\\begin", 
"\\beta", 
"\\bf", 
"\\bibitem", 
"\\bibliography", 
"\\bibliographystyle", 
"\\bigcirc",
"\\bigtriangledown", 
"\\bigtriangleup", 
"\\boldmath", 
"\\bot", 
"\\bowtie",
"\\Box", 
"\\breve", 
"\\bullet", 
"\\bye",
"\\cal", 
"\\cap", 
"\\cdot", 
"\\chapter",
"\\chi", 
"\\circ", 
"\\cite", 
"\\clubsuit",
"\\cong", 
"\\copyright", 
"\\cup", 
"\\dagger", 
"\\dashv", 
"\\date",
"\\ddagger", 
"\\ddot", 
"\\depth",
"\\delta", 
"\\diamond",
"\\diamondsuit", 
"\\div",
"\\documentclass", 
"\\documentstyle", 
"\\dot", 
"\\doteq", 
"\\downarrow",
"\\ell", 
"\\em", 
"\\emph",
"\\emptyset", 
"\\end", 					
"\\enlargethispage",
"\\ensuremath",
"\\epsilon", 
"\\eqnarray", 
"\\equiv",
"\\eta", 
"\\evensidemargin",
"\\exists", 
"\\fbox",
"\\fi",
"\\flat", 
"\\fntguide",
"\\forall",
"\\framebox",
"\\frontmatter", 
"\\frown", 
"\\gamma", 
"\\geq", 
"\\gg", 
"\\gnat"
"\\grave", 
"\\hat", 
"\\hbar", 
"\\headheight",
"\\headings", 
"\\headsep", 
"\\heartsuit", 
"\\height",
"\\hookleftarrow", 
"\\hookrightarrow", 
"\\iffalse",
"\\imath", 
"\\infty", 
"\\iota", 
"\\item", 
"\\itshape",
"\\jmath", 
"\\kappa", 
"\\label",
"\\lambda", 
"\\langle", 
"\\lceil", 
"\\leftarrow",
"\\leftharpoondown",
"\\leftharpoonup", 
"\\leq", 
"\\lfloor",
"\\lhd", 
"\\ll", 
"\\ln",
"\\log",
"\\longleftarrow", 
"\\longleftrightarrow", 
"\\longmapsto", 
"\\longrightarrow",
"\\m",
"\\mainmatter",
"\\makebox",
"\\maketitle", 
"\\mathbf",
"\\mapsto", 
"\\max", 
"\\mbox",
"\\mid", 
"\\models", 
"\\mp", 
"\\mu", 
"\\nabla", 
"\\nearrow",
"\\neg", 
"\\neq",
"\\newblock",
"\\newcommand",
"\\newdimen",
"\\newenvironment",
"\\newmathalphabet",
"\\newlength",
"\\newpage",
"\\newsavebox",
"\\noindent",
"\\nstex",
"\\nu",
"\\nwarrow", 
"\\o", 
"\\oarg",
"\\oddsidemargin", 
"\\odot", 
"\\omega",
"\\ominus", 
"\\oplus", 
"\\oslash", 
"\\otimes", 
"\\pagebreak", 
"\\pagenumbering", 
"\\pagestyle", 
"\\parallel",
"\\part",
"\\paragraph", 
"\\parbox",
"\\perp", 
"\\phi", 
"\\pi", 
"\\pm", 
"\\prec", 
"\\preceq",
"\\prime", 
"\\psi", 
"\\put",
"\\quad",
"\\rangle", 
"\\rceil", 
"\\reasonsection",
"\\rfloor", 
"\\rhd", 
"\\rho", 
"\\rightarrow", 
"\\rightharpoondown", 
"\\rightharpoonup",
"\\rm", 
"\\samepage",
"\\savebox",
"\\sbox",
"\\sc",
"\\searrow", 
"\\section",
"\\setcounter",
"\\settodepth",
"\\settoheight", 
"\\setlength", 
"\\setminus", 
"\\settowidth",
"\\sf", 
"\\sharp", 
"\\sigma", 
"\\sim", 
"\\simeq", 
"\\sl", 
"\\sloppy", 
"\\smile", 
"\\spadsuit", 
"\\sqcap",
"\\sqcup", 
"\\sqsubset", 
"\\sqsubseteq",
"\\sqsupset", 
"\\sqsupseteq", 
"\\star",
"\\stringwidth",
"\\subparagraph",
"\\subsection", 
"\\subset", 
"\\subseteq", 
"\\subsubsection",
"\\succ", 
"\\succeq", 
"\\sum", 
"\\supset",
"\\supseteq", 
"\\surd", 
"\\swarrow", 
"\\tableofcontents",
"\\tabularnewline",
"\\tau",
"\\textasciicircum",
"\\textasciitilde",
"\\textbackslash",
"\\textcircled",
"\\textcompwordmark",
"\\textbar", 
"\\textbf",
"\\textemdash",
"\\textendash",
"\\textexclamdown",
"\\textgreater",
"\\textheight",
"\\textit",
"\\textless",
"\\textmd",
"\\textquestiondown",
"\\textquotedblleft",
"\\textquotedblright",
"\\textquoteleft",
"\\textquoteright",
"\\texttrademark",
"\\textregistered",
"\\textrm",
"\\textsc",
"\\textsf",
"\\textsl",
"\\texttt",
"\\textup",
"\\textwidth", 
"\\thebibliography", 
"\\theta", 
"\\thicklines",
"\\thinlines",
"\\thispagestyle",
"\\tilde", 
"\\times",
"\\title", 
"\\topmargin", 
"\\totalheight",
"\\triangle", 
"\\triangleleft",
"\\triangleright", 
"\\tt", 
"\\unlhd", 
"\\unrhd",
"\\uparrow", 
"\\updownarrow", 
"\\upsilon",
"\\usepackage", 
"\\varepsilon",
"\\varpi",
"\\varrho", 
"\\varsigma", 
"\\vartheta", 
"\\vdash", 
"\\vec", 
"\\vee", 
"\\verb",
"\\usebox",
"\\wedge",
"\\width", 
"\\wp", 
"\\wr", 
"\\xi", 
"\\zeta",
NIL,
};



/* private functions */
static long _tex_in_comment( long offset );
BOOL tex_is_string_token( long* start, long* end );
BOOL tex_is_number_token( long* start, long* end);
BOOL tex_is_keyword_token( int* table, long* start, long* end );

/* public functions */
void DLL _tex_init( void );
BOOL DLL _tex_expand( void );
long DLL _tex_language_colors_update( long firstline, long lastline );
void DLL _tex_language_colors( long numlines );
void DLL _tex_assign_template( LPSTR abbrev, LPSTR template );
void DLL _tex_add_keyword( LPSTR keyword, int table );



/*******************************************************************
 * _init
 *
 * PARAMETERS:  none.
 *
 * DESCRIPTION: Initialization function.  Called when dll is loaded.
 *              Calls the intialization function for this module. 
 *              
 * RETURN:      None.
 *******************************************************************/
void DLL 
_init()
{
    _tex_init();
}


/*******************************************************************
 *  _tex_init()
 *
 *  PARAMETERS:  none.
 *
 *  DESCRIPTION: Initialize this module by exporting any required
 *   functions and creating the hash tables needed to support 
 *   TEX language template expansion and TEX keyword coloring.
 *
 *  RETURN:      none.
 * 
 *  NOTE:        It is recommended that the hash table length be a
 *   prime number to maximize utilization of the hash table.
 ******************************************************************/
void DLL
_tex_init( void )
{
    static BOOL firstTime = TRUE;
    int    i;
	LPSTR hashStr;

    if (firstTime)
    {
        LibExport( "BOOL _tex_expand");
        LibExport( "_tex_assign_template LPSTR LPSTR");
        LibExport( "_tex_add_keyword LPSTR int");
        LibExport( "long _tex_language_colors_update long long");
        LibExport( "_tex_language_colors long");
        LibExport( "_tex_hash_size int int" );
        LibExport( "_tex_routines void" );



		// setup for Outine
		ExtAddOutlineParser("tex", OUTLINE_TYPE_FUNCTION, "Section", 
		  PARSER_FLAG_ALPHA|PARSER_FLAG_NUMERIC|PARSER_FLAG_UNDERSCORE|PARSER_FLAG_ROUTINE,
 		  " {}*\\-\\'\\?!",
 		  "(^\\\\\\c.*section[\\*]*{.*})|^(\\\\\\cpart{.*})|^(\\\\\\c.*paragraph{.*})");
	   
		ExtAddOutlineParser("tex", OUTLINE_TYPE_CLASS, "Class", 
		  PARSER_FLAG_ALPHA|PARSER_FLAG_NUMERIC|PARSER_FLAG_UNDERSCORE|PARSER_FLAG_ROUTINE,
 		  " {}-\\?\\'!","^\\\\\\cdocumentclass{.*}");

		ExtAddOutlineParser("tex", OUTLINE_TYPE_PROCEDURE, "Block", 
		  PARSER_FLAG_ALPHA|PARSER_FLAG_NUMERIC|PARSER_FLAG_UNDERSCORE|PARSER_FLAG_ROUTINE,
 	 	  " {}-\\?\\'!","^\\\\\\cbegin{.*}");

        firstTime = FALSE;

        /* Create a hash table to hold all of the 'TEX' keyword names.
         */
        _tex_hash_names = HashCreateTable( 701/* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _tex_names[ i ] != NIL )
        {
            LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _tex_hash_names, 
                _tex_names[i], strlen( _tex_names[i] ) );
            if (lpf)
            {
                *lpf = StrNew( _tex_names[i] );
                i++;
            }
        }
		hashStr = HashFindFirst(_tex_hash_names, NIL);
		while (hashStr) 
		    hashStr = HashFindNext();


        /* Create a hash table to hold all of the 'TEX' template names.
         */
        _tex_hash_templates = HashCreateTable( 301 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _tex_templates[ i ].abbrev != NIL )
        {
            _tex_assign_template( _tex_templates[i].abbrev, _tex_templates[i].template );
            i++;
        }
        // Setup the %10 template macro to hold the double backslash characters
        ExtSetTemplateMacro( 10, "\\\\", NIL );

    }
}


/*******************************************************************
 * _tex_hash_size()
 *******************************************************************/
void DLL
_tex_hash_size( int ht, int size )
{
    if (size > 0)
    {
        if (ht == 1)
            _tex_hash_size_1 = size;
        else if (ht == 2)
            _tex_hash_size_2 = size;
    }
}

/*******************************************************************
 ** _tex_expand()
 *
 *  DESCRIPTION: Attempts to perform 'TEX' language template 
 *    expansion. This function is normally bound to the <SPACE> key.
 *    When invoked, it reads the previous word.  
 *    If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters 
 *    in the templates have special meanings. They are:
 *
 *      \n  - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 *
 *  PARMETERS:  none.
 *
 *  RETURN:   TRUE if keyword match, FALSE if not a keyword. 
 *
 *  NOTE:     
 *******************************************************************/
BOOL DLL
_tex_expand( void )
{

    LPVOID htEntry;
    LPSTR  _tex_string = NULL;
    LPSTR  str;

    // Save the current buffer position on the stack
    MarkSavePos();

    // Use spaces - disable the use of tab characters
    BufSetTabUsage( FALSE );

    // Check to see if we are currently positioned inside a 'TEX' comment
    if (_tex_in_comment( -1L ) != -1)
        goto _tex_expand_incomplete;
        

    // Allocate a buffer and read entire line of starting cursor position
    // The maximum string length returned by BufReadStr is 32767 chars 
    str = BufReadStr( 65535 );     

    // Strip line to first word 
    _tex_string = StrTrim( str, " \t" );
    
    if (!_tex_string)
        goto _tex_expand_incomplete;

    /* Search for a keyword match in the hash table.*/
    htEntry = HashFindEntry( _tex_hash_templates, _tex_string, strlen( _tex_string ) );
    
    // delete the buffer 
    StrFree( _tex_string );

    if (htEntry)
    {
        /* A template exists for the current keyword, delete the keyword
         * and insert the template.
         */
        // Remove the last saved position from the stack
        MarkDropPos( 1 );
        
        str = *(LPSTR*)htEntry;
    
        // delete the word
        BufDelToEOL();

        // replace the word with the expanded template
        _ext_expand_template( str );
    } 
    else   // no template found for the current word.
        goto _tex_expand_incomplete;
    
    return TRUE;

_tex_expand_incomplete:
    
    // restore position saved on entry
    MarkRestorePos();

    return FALSE;

}
/*
 ** _tex_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_tex_language_colors_update( long firstline, long lastline )
{
	long startoff;
   	long count = 0;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);
		startoff = _PosQOffset();

		/* Clear all attributes for the line range and
		 * the reset them within _tex_language_colors.
		 */
		AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
	   _PosInit( startoff );
		count = lastline - firstline + 1;
		_tex_language_colors( lastline - firstline + 1 );
	}
	return count;
}


/*****************************************************************
 * _tex_language_colors()
 *
 * PARAMETERS:  long numlines - number of lines to process
 *
 * DESCRIPTION: Sets a block of text according to TEX language
 *   syntax.  If the given number of lines equals (-1), coloring
 *   begins at the top of the buffer, otherwise coloring begins
 *   at the current _Pos position.
 *
 *   Each line of text is processed.  First the line is checked for
 *   an '*' or '\' in column 7.  If found the entire line is colored
 *   as a comment and we move on.  Next columns 0-6 are colored the
 *   keyword color.  The remainder of the line is parsed one token at
 *   a time and colored according to its token type, string, number
 *   or TEX language keyword.  The two variables start and end are
 *   maintained while each line is parsed.  Start holds the _Pos offset
 *   of the first character of the token.  End maintains the _Pos offset
 *   one past the last character of the token.
 *
 * RETURN:      none.
 * 
 * NOTE:        It is assumed that if numlines != -1, that _PosInit()
 *   has been called and numlines is relative to the current 
 *   position.
 *
 * TODO:        check additional keyword tables.
 ******************************************************************/
void DLL
_tex_language_colors( long numlines )
{ 
	long  start;
	//long  startline;
    //    long  startoff;
	long  end;
    int table_index = 0;     // table of keywords to check
    
	long  inc;
	WORD  ch;
   	int   i = 0;
	int   comment_color = ColorComments( -1 );
	int   keyword_color = ColorKeywords( -1 );
    int   keyword_color_1 = ColorAlternate5( -1 );
    int   keyword_color_2 = ColorAlternate6( -1 );
    
	long  maxlines = ExtCommentSearchLimit( -1 );
	int   quote_color      = ColorAlternate1( -1 );
	int   number_color     = ColorAlternate2( -1 );
	int   brace_color      = ColorAlternate3( -1 );
	int   preproc_color    = ColorAlternate4( -1 );

   	LPSTR buf;
    
    buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	
	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 ); 
		start = 0; 
		numlines = 0x7FFFFFFF;
	}
	else
	{
		start  = _PosQOffset();
	}
   	end = start + 1;
	ch = _PosCurrentCharNL();

	start = _PosQOffset();

	maxlines = max( numlines, maxlines );

    for(;( ch != EOF_CHAR) && (numlines > 0); numlines-- )
	{
		// check for a comment
		if (ch == '%')
		{
			start = _PosQOffset();

			inc = _PosNextLine( 1 );
		   	end = _PosQOffset();
			_PosInit( start );
		   _PosSetColor( comment_color, end - start - inc);

			_PosInit( end );
			ch = _PosCurrentCharNL();
			continue;
		}
		else
        {
            while ( !(end_of_line(ch) ) ) 
            {                 
               
                if ( tex_is_keyword_token( &table_index , &start, &end ) )
                {
                    if( table_index == 0 )
                    {
                       _PosSetColor( keyword_color, end - start );
                    }
                    else if( table_index == 1 )
                    {
                    
                       _PosSetColor( keyword_color_1, end - start );
                    }                   
                    else if( table_index == 2 )
                    {
                       _PosSetColor( keyword_color_2, end - start );
                    }
                }// end if keyword
				else if (brace_color && _fstrchr("{}[]()", ch) != 0)    
				{                                        
					long coff;                            

					coff = _PosQOffset();            
					_PosInit(coff);                  
					_PosSetColor(brace_color, 1);    
					_PosInit(coff);                  
				}               
   			    else if (tex_is_string_token( &start, &end ) ) 
                {                       
                    _PosSetColor( quote_color, end - start );
                } 
                else if ( tex_is_number_token( &start, &end ) ) 
                {
                    _PosSetColor( number_color, end - start );
                }    

                start = end;
                _PosInit( start );           

                // skip leading white space
                for ( ch = _PosCurrentCharNL(); ch == ' ' ; ch = _PosNextCharNL(), start++ );
                    end = start + 1; 

  				if (ch == EOL2_CHAR)
					end++;

            } // end while not EOL
        } // end if not comment
       // -cb if ( end_of_line(ch) )

        if ( end_of_line(ch) )
	    {            
            ch = _PosNextCharNL();
        }             
    } // end while !EOF and more lines to do
      
   MemFree( buf );
}

                                   

/*
 ** _tex_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'TEX' comment by search backward for a % 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_tex_in_comment( long offset )
{
	WORD  ch;

	/* Check to see if we are currently positioned inside a 'TEX' comment
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	do
	{
		if (ch == '\n')
			return -1;
		else if (ch == '%')
			return _PosQOffset();  /* currrent line is in a comment */
		ch = _PosPrevChar();
	} while (ch != EOF_CHAR);

	return -1L;
}


/*******************************************************************
 * tex_is_string_token( long* start, long *end )
 *
 * PARAMETERS:  long* start - starting offset of string if found
 *              long* end   - ending offset if found
 *
 * DESCRIPTION: Determines if current buffer position is within a
 *  quoted string and the start and end offset of the string.
*******************************************************************/
BOOL 
tex_is_string_token( long* start, long *end )
{
    BOOL result = FALSE;  
    long offset_to_restore;
    long offset;
    WORD ch;
    
    // save the pos position
    offset_to_restore = _PosQOffset();
    offset = offset_to_restore;
    ch = _PosCurrentCharNL();

    // check for string
    if( ch == '"')
    {
        *start = offset;
        result = TRUE;
        ch = _PosNextCharNL();
        while ( ch != '\'' && ch != '"' && !(end_of_line(ch)) )
        {
            ch = _PosNextCharNL();
        } 
        *end = _PosQOffset() + 1;
    } // end if string found
   
    _PosInit( offset_to_restore );
    return( result );
}


/*******************************************************************
 * tex_is_number_token( long* start, long *end )
 *
 * PARAMETERS:  long* start - starting offset of number if found
 *              long* end   - ending offset if found
 *
 * DESCRIPTION: Determines if current buffer position is a
 *   number and the start and end offset of the number.
 *
 *******************************************************************/
BOOL 
tex_is_number_token( long* start, long *end)
{     
    
    WORD ch;
    long offset_to_restore;
    
    BOOL is_number = FALSE;
    BOOL period_found = FALSE;
    BOOL leading_punct = FALSE;
    BOOL done =FALSE;
                                       
                                       
    // save the pos position
    offset_to_restore = _PosQOffset();
      
    // get the first character
    ch = _PosCurrentCharNL();

    // note if leading .
    if ( ch == '.' )
    {
        period_found = TRUE;
        ch = _PosNextCharNL();
    } 
     
    // note if leading valid delimiter 
    if ( ch == '{' || ch == '(' )
    {
        ch = _PosNextCharNL(); 
        leading_punct = TRUE;
    } 
     
    if ( !isdigit((BYTE)ch) )
    {
        done = TRUE;
    }

    while( !done && !( end_of_line(ch) ))
    {
        ch = _PosNextCharNL();
        if ( ch == '.')
        {
            // if a second period - it's not a valid number
            if ( period_found)
            {
                 done = TRUE;
            }
            period_found = TRUE;
        }            
        else if( !(isdigit((BYTE)ch) )  )
        {
            done = TRUE;

            // if valid delimiter
            if( ch == '}' || ch == ')' ||  ch == ' ' || ch == '\t'
               ||  end_of_line(ch) )
            {
                is_number = TRUE;
                *start = offset_to_restore;  
                *end = _PosQOffset(); 
  
                if( leading_punct ) 
                {
                    offset_to_restore++; 
                    (*start)++ ;
                }                     
                
            } // end if valid delimiter
        } // end non digit found following digit
    } // end while !done
 
     // restore the pos position
     
    _PosInit( offset_to_restore );
    return( is_number );
}


/*******************************************************************
 * tex_is_keyword_token( int table_index, long* start, long *end )
 *
 * PARAMETERS:  LPVOID table - table of keywords
 *              long* start  - starting offset of keyword if found
 *              long* end    - ending offset if found
 *
 * DESCRIPTION: Determines if current buffer position is within a
 *  quoted keyword and the start and end offset of the keyword.
 *
 *  TEX keyword rules.
 *   Keywords are defined in hash tables.
 *   _tex_hash_names hash table loaded from tex_names array
 *   _tex_hash_names1 hash table is created and loaded at runtime
 *   _tex_hash_names2 hash table is created and loaded at runtime
 *   See HashCreateTable. 
 *
 *  RETURN: TRUE if in a keyword, FALSE if not in a keyword.
 *  Sets given start and end parameters to starting offset of keyword
 *  and ending offset of keyword.
 *
 *  NOTES:
 *******************************************************************/
BOOL 
tex_is_keyword_token( int *table, long* start, long* end )
{
    BOOL keyword_found = FALSE;
    LPSTR lp;
    long offset_to_restore;
    int length;
    WORD ch; 

    // save the pos position
    offset_to_restore = _PosQOffset();

    // set ptr to global buffer
    lp = buf;
                     
    length = 0;                 
	for ( ch = _PosCurrentCharNL(); !(_fstrchr("_$[]{}^| \t*", ch))
      && !end_of_line(ch) && length < MAX_BUF_LEN;  ch = _PosNextCharNL() ) 
                   
       {
           *lp++ = (char)ch;
           length++;
       }
    *lp = 0;
 
    if ( length > 0 && _tex_hash_names  )
    {
        if (HashFindEntry( _tex_hash_names, buf, length ) )
        {
           keyword_found = TRUE;
           *table = 0; 
        }
    }              
    if ( length > 0 &&  _tex_hash_names_1 && !keyword_found  )
    {
        if (HashFindEntry( _tex_hash_names_1, buf, length) )
        {
           keyword_found = TRUE;
           *table = 1;
        }
    }
    
    if ( length > 0 &&  _tex_hash_names_2 && !keyword_found )
    {
        if (HashFindEntry( _tex_hash_names_2, buf, length) )
        {
            keyword_found = TRUE;
            *table = 2;
        }
    }
    *start = offset_to_restore; 
    
    if ( length < 1 )
    {
        *end = offset_to_restore + 1;
    }                           
    else
    {
        *end = offset_to_restore + length;
    }
    
    _PosInit( offset_to_restore );
    return( keyword_found ); 
}



/*******************************************************************
 * _tex_add_keyword
 *
 * PARAMETERS:  LPSTR keyword - keyword to add to the table
 *              int table     - table to add the keyword to.
 *
 * DESCRIPTION: Adds a keyword to a hash table
 *
 * RETURN:      none.
 *              
 * NOTES:
 *******************************************************************/
void DLL
_tex_add_keyword( LPSTR keyword, int table )
{
    if (keyword && *keyword && table >= 0)
    {
        LPSTR XFAR *lpf;
        LPVOID  ht;

        switch( table )
        {
        case 0:
            ht = _tex_hash_names;
            break;
        case 1:
            if (_tex_hash_names_1 == NULL)
                _tex_hash_names_1 = HashCreateTable( _tex_hash_size_1, sizeof( LPSTR ), 0 );
            ht = _tex_hash_names_1;
            break;
        case 2:
            if (_tex_hash_names_2 == NULL)
                _tex_hash_names_2 = HashCreateTable( _tex_hash_size_2, sizeof( LPSTR ), 0 );
            ht = _tex_hash_names_2;
            break;
        }

        lpf = (LPSTR XFAR *)HashGetEntry( ht, 
                                keyword, strlen( keyword ) );
        if (lpf)
            *lpf = StrNew( keyword );
    }
}


/*******************************************************************
 * _tex_assign_template()
 *
 * PARAMETERS:  abbrev   - keyword abreviation used to locate a template
 *              template - encoded template string associated with 'abbrev'
 *
 * DESCRIPTION: Add or modify a 'TEX' template string at runtime.
 *
 * RETURN:      None.
 *
 * NOTES:
 *******************************************************************/
void DLL
_tex_assign_template( LPSTR abbrev, LPSTR template )
{
    _ext_assign_template( _tex_hash_templates, abbrev, template );
}


/*--  Private functions --*/
  
/*******************************************************************
 * _tex_routines()
 *
 * PARAMETERS:  None.
 *
 * DESCRIPTION:
 *
 * RETURN:
 *
 * NOTES:
 *******************************************************************/
void DLL
_tex_routines( void )
{
	if (Grep("\\\\part|\\\\subparagragh|\\\\paragraph|\\\\section|\\\\subsection|\\\\subsubsection|\\\\reasonsection"))
		MsgMessage("");
	else
		MsgMessage("No routines found.");
}
    

